local Heatsink = LibStub("AceAddon-3.0"):NewAddon("Heatsink", "AceEvent-3.0", "AceTimer-3.0")
local L = LibStub:GetLibrary("AceLocale-3.0"):GetLocale("Heatsink")
local candy = LibStub("LibCandyBar-3.0")
local icd = LibStub("LibInternalCooldowns-1.0")
local media = LibStub("LibSharedMedia-3.0")
local db, anchor, pguid, pclass

local BOOKTYPE_PET = _G.BOOKTYPE_PET
local BOOKTYPE_SPELL = _G.BOOKTYPE_SPELL
local CreateFrame = _G.CreateFrame
local GameFontNormal = _G.GameFontNormal
local GetContainerItemCooldown = _G.GetContainerItemCooldown
local GetContainerItemInfo = _G.GetContainerItemInfo
local GetContainerItemLink = _G.GetContainerItemLink
local GetContainerNumSlots = _G.GetContainerNumSlots
local GetInventoryItemCooldown = _G.GetInventoryItemCooldown
local GetInventoryItemLink = _G.GetInventoryItemLink
local GetInventoryItemTexture = _G.GetInventoryItemTexture
local GetInventorySlotInfo = _G.GetInventorySlotInfo
local GetItemInfo = _G.GetItemInfo
local GetNumSpellTabs = _G.GetNumSpellTabs
local GetRealZoneText = _G.GetRealZoneText
local GetSpellCooldown = _G.GetSpellCooldown
local GetSpellInfo = _G.GetSpellInfo
local GetSpellName = _G.GetSpellName
local GetSpellTabInfo = _G.GetSpellTabInfo
local GetTime = _G.GetTime
local IsActiveBattlefieldArena = _G.IsActiveBattlefieldArena
local UnitClass = _G.UnitClass
local UnitGUID = _G.UnitGUID
local ipairs = _G.ipairs
local math = _G.math
local pairs = _G.pairs
local select = _G.select
local string = _G.string
local table = _G.table
local unpack = _G.unpack
local wipe = _G.wipe

local defaults = {
	profile = {
		min = 3,
		max = 3600,
		growup = true,
		texture = "Blizzard",
		font = "ABF",
		width = 250,
		scale = 1,
		pos = {
			p = "CENTER",
			rp = "CENTER",
			x = 0,
			y = 0,
		},
		color = {
			bg = { 0.5, 0.5, 0.5, 0.3 },
			text = { 1, 1, 1 },
			bar = { 0.25, 0.33, 0.68, 1 },
		},
		show = {
			school = true,
			spells = true,
			pet = true,
			items = true,
			inventory = true,
			proc = true,
		},
	},
}

-- Credit to the BigWigs team (Rabbit, Ammo, et al) for the anchor code 
local function sortBars(a, b)
	return (a.remaining > b.remaining and db.growup) and true or false
end

local function rearrangeBars(anchor)
	local tmp = {}
	for bar in pairs(anchor.active) do
		table.insert(tmp, bar)
	end
	table.sort(tmp, sortBars)
	local lastBar = nil
	for i, bar in ipairs(tmp) do
		bar:ClearAllPoints()
		if db.growup then
			bar:SetPoint("BOTTOMLEFT", lastBar or anchor, "TOPLEFT")
			bar:SetPoint("BOTTOMRIGHT", lastBar or anchor, "TOPRIGHT")
		else
			bar:SetPoint("TOPLEFT", lastBar or anchor, "BOTTOMLEFT")
			bar:SetPoint("TOPRIGHT", lastBar or anchor, "BOTTOMRIGHT")
		end
		lastBar = bar
	end
	wipe(tmp)
end

local function onDragHandleMouseDown(self)
	self:GetParent():StartSizing("BOTTOMRIGHT")
end

local function onDragHandleMouseUp(self, button)
	self:GetParent():StopMovingOrSizing()
end

local function onResize(self, width)
	db.width = width
	rearrangeBars(self)
end

local function onDragStart(self)
	self:StartMoving()
end

local function onDragStop(self)
	self:StopMovingOrSizing()
	local p, _, rp, x, y = self:GetPoint()
	db.pos.p = p
	db.pos.rp = rp
	db.pos.x = x
	db.pos.y = y
end

local function onControlEnter(self)
	GameTooltip:ClearLines()
	GameTooltip:SetOwner(self, "ANCHOR_TOPLEFT")
	GameTooltip:AddLine(self.tooltipHeader)
	GameTooltip:AddLine(self.tooltipText, 1, 1, 1, 1)
	GameTooltip:Show()
end

local function onControlLeave()
	GameTooltip:Hide()
end

local function getBar(text)
	local bar = nil
	for k in pairs(anchor.active) do
		if k.candyBarLabel:GetText() == text then
			bar = true
		end
	end
	return bar
end

local function stopBar(text)
	local bar = nil
	for k in pairs(anchor.active) do
		if (not text or k.candyBarLabel:GetText() == text) then
			k:Stop()
			bar = true
		end
	end
	if bar then rearrangeBars(anchor) end
	return bar
end

local function startBar(text, start, duration, icon, force)
	if not force and getBar(text) then return end
	local remaining
	if start then
		remaining = duration - (GetTime() - start)
	end
	stopBar(text)
	local bar = candy:New(media:Fetch("statusbar", db.texture), 200, 14)
	bar:Set("anchor", anchor)
	anchor.active[bar] = true
	bar.candyBarBackground:SetVertexColor(unpack(db.color.bg))
	bar:SetColor(unpack(db.color.bar))
	bar.candyBarLabel:SetTextColor(unpack(db.color.text))
	bar.candyBarLabel:SetFont(media:Fetch("font", db.font), 10)
	bar.candyBarDuration:SetFont(media:Fetch("font", db.font), 10)
	bar:SetLabel(text)
	bar:SetDuration(remaining or duration)
	bar:SetTimeVisibility(true)
	bar:SetIcon(icon)
	bar:SetScale(db.scale)
	bar:Start()
	rearrangeBars(anchor)
end

local function runTest(anchor)
	local tmp = {}
	for i = 2, GetNumSpellTabs() do
		local name, texture, offset, numSpells = GetSpellTabInfo(i)
		if offset then
			for s = offset + 1, offset + numSpells do
				table.insert(tmp, (GetSpellName(s, BOOKTYPE_SPELL)))
			end
		end
	end
	local spell = tmp[math.random(1, #tmp)]
	local name, rank, icon = GetSpellInfo(spell)
	local duration = math.random(10, 30)
	startBar(name, nil, duration, icon)
	wipe(tmp)
end

local function toggleAnchor(anchor)
	if anchor:IsShown() then
		anchor:Hide()
	else
		anchor:Show()
	end
end

local function createAnchor(frameName, title)
	local display = CreateFrame("Frame", frameName, UIParent)
	display:EnableMouse(true)
	display:SetMovable(true)
	display:SetResizable(true)
	display:RegisterForDrag("LeftButton")
	display:SetWidth(db.width or 200)
	display:SetHeight(20)
	display:SetMinResize(80, 20)
	display:SetMaxResize(1920, 20)
	display:ClearAllPoints()
	display:SetPoint(db.pos.p, UIParent, db.pos.rp, db.pos.x, db.pos.y)

	local bg = display:CreateTexture(nil, "PARENT")
	bg:SetAllPoints(display)
	bg:SetBlendMode("BLEND")
	bg:SetTexture(0, 0, 0, 0.3)

	local header = display:CreateFontString(nil, "OVERLAY")
	header:SetFontObject(GameFontNormal)
	header:SetText(title)
	header:SetAllPoints(display)
	header:SetJustifyH("CENTER")
	header:SetJustifyV("MIDDLE")

	local drag = CreateFrame("Frame", nil, display)
	drag:SetFrameLevel(display:GetFrameLevel() + 10)
	drag:SetWidth(16)
	drag:SetHeight(16)
	drag:SetPoint("BOTTOMRIGHT", display, -1, 1)
	drag:EnableMouse(true)
	drag:SetScript("OnMouseDown", onDragHandleMouseDown)
	drag:SetScript("OnMouseUp", onDragHandleMouseUp)
	drag:SetAlpha(0.5)

	local tex = drag:CreateTexture(nil, "BACKGROUND")
	tex:SetTexture("Interface\\AddOns\\Heatsink\\Textures\\draghandle")
	tex:SetWidth(16)
	tex:SetHeight(16)
	tex:SetBlendMode("ADD")
	tex:SetPoint("CENTER", drag)

	local test = CreateFrame("Button", nil, display)
	test:SetPoint("BOTTOMLEFT", display, "BOTTOMLEFT", 3, 3)
	test:SetHeight(14)
	test:SetWidth(14)
	test.tooltipHeader = L["Test"]
	test.tooltipText = L["Creates a new test bar"]
	test:SetScript("OnEnter", onControlEnter)
	test:SetScript("OnLeave", onControlLeave)
	test:SetScript("OnClick", function() runTest() end)
	test:SetNormalTexture("Interface\\AddOns\\Heatsink\\Textures\\test")

	local close = CreateFrame("Button", nil, display)
	close:SetPoint("BOTTOMLEFT", test, "BOTTOMRIGHT", 4, 0)
	close:SetHeight(14)
	close:SetWidth(14)
	close.tooltipHeader = L["Hide"]
	close.tooltipText = L["Hides the anchor"]
	close:SetScript("OnEnter", onControlEnter)
	close:SetScript("OnLeave", onControlLeave)
	close:SetScript("OnClick", function() toggleAnchor(anchor) end)
	close:SetNormalTexture("Interface\\AddOns\\Heatsink\\Textures\\close")

	display:SetScript("OnSizeChanged", onResize)
	display:SetScript("OnDragStart", onDragStart)
	display:SetScript("OnDragStop", onDragStop)
	display.active = {}
	display:Hide()
	return display
end

local function updateAnchor(anchor)
	for bar in pairs(anchor.active) do
		bar.candyBarBar:SetStatusBarTexture(media:Fetch("statusbar", db.texture))
		bar.candyBarBackground:SetTexture(media:Fetch("statusbar", db.texture))
		bar.candyBarBackground:SetVertexColor(unpack(db.color.bg))
		bar.candyBarBar:SetStatusBarColor(unpack(db.color.bar))
		bar.candyBarLabel:SetTextColor(unpack(db.color.text))
		bar.candyBarLabel:SetFont(media:Fetch("font", db.font), 10)
		bar.candyBarDuration:SetFont(media:Fetch("font", db.font), 10)
		bar:SetScale(db.scale)
	end	
	rearrangeBars(anchor)
end

function Heatsink:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("HeatsinkDB", defaults, "Default")
	db = self.db.profile
	self.db.RegisterCallback(self, "OnProfileChanged", "UpdateProfile")
	self.db.RegisterCallback(self, "OnProfileCopied", "UpdateProfile")
	self.db.RegisterCallback(self, "OnProfileReset", "UpdateProfile")
	anchor = createAnchor("HeatsinkAnchor", "Heatsink")
	candy.RegisterCallback(self, "LibCandyBar_Stop")
	self:SetupOptions()
end

function Heatsink:OnEnable()
	self:RegisterEvent("BAG_UPDATE_COOLDOWN")
	self:RegisterEvent("PET_BAR_UPDATE_COOLDOWN")
	self:RegisterEvent("SPELLS_CHANGED")
	self:RegisterEvent("SPELL_UPDATE_COOLDOWN")
	self:RegisterEvent("UNIT_INVENTORY_CHANGED")
	self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA")

	icd.RegisterCallback(self, "InternalCooldowns_Proc")

	pguid = pguid or UnitGUID("player")
	pclass = pclass or select(2, UnitClass("player"))

	self:ScanSpells()
	self:ScanPetSpells()
	self:ScanItems()
	self:ScanInventory()

	self:BAG_UPDATE_COOLDOWN()
	self:PET_BAR_UPDATE_COOLDOWN()
	self:SPELL_UPDATE_COOLDOWN()
	self:UNIT_INVENTORY_CHANGED()
end

function Heatsink:OnDisable()
	self:UnregisterAllEvents()
	icd.UnregisterCallback(self, "InternalCooldowns_Proc")
end

function Heatsink:UpdateProfile()
	db = self.db.profile
	updateAnchor(anchor)
end

function Heatsink:RunTest()
	runTest(anchor)
end

function Heatsink:ToggleAnchor()
	toggleAnchor(anchor)
end

function Heatsink:UpdateAnchor()
	updateAnchor(anchor)
end

function Heatsink:LibCandyBar_Stop(callback, bar)
	local a = bar:Get("anchor")
	if a == anchor and anchor.active and anchor.active[bar] then
		anchor.active[bar] = nil
		--rearrangeBars(anchor)
	end
end

function Heatsink:InternalCooldowns_Proc(callback, item, spell, start, duration, source)
	if self.db.profile.show.proc then
		local name, icon, _
		if source == "ENCHANT" then
			name, _, icon = GetSpellInfo(spell)
		else
			name, _, _, _, _, _, _, _, _, icon = GetItemInfo(item)
		end
		startBar(name, start, duration, icon)
	end
end

function Heatsink:SPELLS_CHANGED()
	self:ScanSpells()
	self:ScanPetSpells()

	self:SPELL_UPDATE_COOLDOWN()
	self:PET_BAR_UPDATE_COOLDOWN()
end

function Heatsink:SPELL_UPDATE_COOLDOWN()
	if self.db.profile.show.spells then
		if self.db.profile.show.school and Heatsink.baseline[pclass] then
			for school, spell in pairs(Heatsink.baseline[pclass]) do
				local start, duration, active = GetSpellCooldown(spell)
				if active == 1 and duration >= self.db.profile.min and duration <= self.db.profile.max then
					local _,_,icon = GetSpellInfo(spell)
					startBar(string.format(L["%s School"],school), start, duration, icon)
				end
			end
		end
		for name, spell in pairs(Heatsink.player) do
			if self.db.profile.show.school and spell.school and getBar(string.format(L["%s School"],spell.school)) then
				return
			elseif spell.id then
				local start, duration, active = GetSpellCooldown(spell.id, BOOKTYPE_SPELL)
				if active == 1 and duration >= self.db.profile.min and duration <= self.db.profile.max then
					for k,v in pairs(Heatsink.spellreplace) do
						if name == k then
							name =v
							break
						end
					end
					startBar(name, start, duration, spell.icon)
				end
			end
		end
	end
end

function Heatsink:PET_BAR_UPDATE_COOLDOWN()
	if self.db.profile.show.pet then
		for name, spell in pairs(Heatsink.pet) do
			local start, duration, active = GetSpellCooldown(spell.id, BOOKTYPE_PET)
			if active == 1 and duration >= self.db.profile.min and duration <= self.db.profile.max then
				startBar(name, start, duration, spell.icon)
			end
		end
	end
end

function Heatsink:UNIT_SPELLCAST_SUCCEEDED(event, unit, name)
	if unit == "player" and Heatsink.chain[name] then
		stopBar()
		Heatsink:SPELL_UPDATE_COOLDOWN()
		Heatsink:PET_BAR_UPDATE_COOLDOWN()
		Heatsink:UNIT_INVENTORY_CHANGED()
		Heatsink:BAG_UPDATE_COOLDOWN()
	end
end

function Heatsink:ZONE_CHANGED_NEW_AREA()
	local zone = GetRealZoneText()
	if not zone then
		self:ScheduleTimer(self.ZONE_CHANGED_NEW_AREA, 2, self)
	elseif IsActiveBattlefieldArena() then
		stopBar()
	end
end

function Heatsink:UNIT_INVENTORY_CHANGED(event, unit)
	if unit == "player" then
		self:ScanItems()
		self:ScanInventory()
	end
end

function Heatsink:BAG_UPDATE_COOLDOWN(event, container)
	self:ScanItems()
	self:ScanInventory()
end

function Heatsink:ScanSpells()
	Heatsink.player = Heatsink.player or {}
	wipe(Heatsink.player)
	for i = 1, GetNumSpellTabs() do
		local name, texture, offset, numSpells = GetSpellTabInfo(i)
		for j = 1, numSpells do
			local id = j + offset
			local name, rank, icon = GetSpellInfo(id, BOOKTYPE_SPELL)
			Heatsink.tooltip:SetSpell(id, BOOKTYPE_SPELL)
			for i = 1, Heatsink.tooltip:NumLines() do
				local line = Heatsink.tooltip.R[i] or ""
				if line:find(L["cooldown$"]) then
					Heatsink.player[name] = {
						id = id,
						icon = icon,
					}
					if Heatsink.school[pclass] and Heatsink.school[pclass][name] then
						Heatsink.player[name].school = Heatsink.school[pclass][name]
					end
				end
			end
		end
	end
end

function Heatsink:ScanPetSpells()
	Heatsink.pet = Heatsink.pet or {}
	wipe(Heatsink.pet)
	local id = 1
	local name, rank, icon = GetSpellInfo(id, BOOKTYPE_PET)
	local previous
	while name do
		if name ~= previous then
			Heatsink.tooltip:SetSpell(id, BOOKTYPE_PET)
			for i = 1, Heatsink.tooltip:NumLines() do
				local line = Heatsink.tooltip.R[i] or ""
				if line:find(L["cooldown$"]) then
					Heatsink.pet[name] = {
						id = id,
						icon = icon,
					}
				break
				end
			end
		end
		id = id + 1
		previous = name
		name = GetSpellInfo(id, BOOKTYPE_PET)
	end
end

function Heatsink:ScanItems()
	if self.db.profile.show.items then
		for _, slotName in pairs(Heatsink.slots) do
			local slotId = GetInventorySlotInfo(slotName)
			local start, duration, active = GetInventoryItemCooldown("player", slotId)
			if active == 1 and duration > self.db.profile.min and duration <= self.db.profile.max then
				local name = select(3, GetInventoryItemLink("player", slotId):find("Hitem[^|]+|h%[([^[]+)%]"))
				local icon = GetInventoryItemTexture("player", slotId)
				local force = false
				if ((slotName == "Trinket0Slot" or slotName == "Trinket1SLot") and (duration < 30)) then
					force = true
				end
				startBar(name, start, duration, icon, force)
			end
		end
	end
end

function Heatsink:ScanInventory()
	if self.db.profile.show.inventory then
		for bag = 0,4 do
			local slots = GetContainerNumSlots(bag)
			for slot = 1,slots do
				local start, duration, active = GetContainerItemCooldown(bag,slot)
				if active == 1 and duration > self.db.profile.min and duration <= self.db.profile.max then
					local name = select(3, GetContainerItemLink(bag,slot):find("Hitem[^|]+|h%[([^[]+)%]"))
					local icon = GetContainerItemInfo(bag, slot)
					for pattern, replacement in pairs(Heatsink.itemreplace) do
						if (string.find(name, pattern)) then
							name = replacement
							break
						end
					end
					startBar(name, start, duration, icon)
				end
			end
		end
	end
end
